/*
** Project: SnmpTool - Simple Network Management Protocol Tool for Win32
** Purpose: SNMP Testing and Diagnostics Utility
** Author:  James D. Murray <jdm@oreilly.com>
** Date:    28 Aug 1997
** File:    SNMPTOOL.C
** Compile: CL SNMPTOOL.C SNMPAPI.LIB MGMTAPI.LIB USER32.LIB
**
** SnmpTool is a work-alike to the Microsoft SNMPUTIL example program.
** SnmpTool implements the complete functionality of SNMPUTIL and adds
** additional features, such as the ability to perform polling, 
** set-request operations, and reading command line arguments from a file.
**
** SnmpTool is a Win32 MS-DOS Console program and uses the Microsoft
** Windows NT SNMP API library (SNMPAPI.DLL) and SNMP Management API
** library (MGMTAPI.DLL). The SNMP Management DLL is not distibuted for
** Windows 95 systems, but the NT DLL will operate correctly when
** installed in the Win95 %SystemRoot%\SYSTEM directory. However, trap
** reception will not function because there is no Microsoft SNMP Trap
** Service for Windows 95.
** 
** SnmpTool is run on the command line in an MS-DOS console window.
** All SnmpTool operations require that the type of operation, the IP
** address or DNS name of the SNMP host, and the read-only or
** read-write community name of the SNMP host be specified on the
** command line. GET, GETNEXT, SET, and POLL operations also require
** one or more OIDs (WALK uses only a single OID). Absolute OIDs must
** begin with a period (i.e. .1.3.6.1.2.1.1.4.0).  OIDs relative to
** MIB-II may be abbreviated and without a leading period (e.g.
** 1.4.0).  The SET operation also requires that each OID also
** specify the data type and value to set.
**
** Prolonged operations, such as polling, MIB walking, and waiting
** for traps, can be halted using the Escape key or Control C. 
**
** SnmpTool operations can be altered using one or more command line
** argument flags. The supported SnmpTool command line arguments are:
**
**   -t <milliseconds>  Number of milliseconds to wait before timing 
**                        out on a request.
**   -r <number>        Number of request retry attempts to perform 
**                        before timing out.
**   -i <milliseconds>  Number of milliseconds to wait between polls.
**   -n <number>        Number of polling requests to perform.
**   -l                 Perform long walk to the end of the MIB view 
**                        (default is walking to the end of the branch).
**   -p                 Use polling for trap detection (default is 
**                        event-driven trap detection).
**   -x                 Use GETNEXT operation for polling (GET operation 
**                        for polling is default).
** 
** Examples of usage:
**
** SnmpTool get 1.2.3.4 public 1.5.0
**   Get the value of the MIB variable 1.3.6.1.2.1.1.5.0
**
** SnmpTool -t 3000 -r 4 get 1.2.3.4 public .1.3.6.1.2.1.1.5.0
**   Get the value of the MIB variable 1.3.6.1.2.1.1.5.0 using a
**   time-out of 3000 milliseconds and four retries.
**
** SnmpTool getnext 1.2.3.4 public 1.4.0
**   Get the next MIB variable value after 1.3.6.1.2.1.1.4.0
**
** SnmpTool -n 200 poll 1.2.3.4 public 1.5.0 1.6.0
**   Poll the MIB variables 1.3.6.1.2.1.1.5.0 and 1.3.6.1.2.1.1.6.0
**   using the default polling interval (3000 milliseconds) and default
**   number of polls (10).
**
** SnmpTool -n 200 -i 60000 poll 1.2.3.4 public .1.3.6.1.2.1.1.3.0
**   Get the value of MIB variable 1.3.6.1.2.1.1.3.0 every 60 seconds 
**   for 200 polls.
**
** SnmpTool walk 1.2.3.4 public .1.3.6.1.4.1.2035.1
**   Perform a GETNEXT walk starting 1.3.6.1.4.1.2035.1 and ending
**   at the start of the next branch (i.e. 1.3.6.1.4.1.2035.2).
**
** SnmpTool -l walk 1.2.3.4 public .1.3.6.1.4.1.77.1
**   Perform a long GETNEXT walk starting 1.3.6.1.4.1.77.1 and 
**   ending at the bottom of the agent's MIB view.
**
** SnmpTool -l walk 1.2.3.4 public .1.3.6.1.4.1.77.1 > walkies.lst 
**   Save results of a long walk to the file walkies.lst.
**
** SnmpTool set 1.2.3.4 private 1.5.0 2 "Fred"
**   Set the sysName (1.3.6.1.2.1.1.5.0) variable value to "Fred".
**
** SnmpTool set 1.2.3.4 private 1.6.0 2 "Phone closet" .1.3.6.1.2.1.2035.1.7.1.1.0 1 42
**   Set the sysLocation (1.3.6.1.2.1.1.6.0) variable value to "Phone closet"
**   and 1.3.6.1.2.1.2035.1.7.1.1.0 to 42.
**
** SnmpTool trap
**   Listen for traps using the event-driven mechanism of the
**   Windows NT SNMP Trap Service.
**
** SnmpTool -p -n 5000 -i 30000 trap
**   Check for traps every 30 seconds for 5000 polls.
**
** SnmpTool trap > traps.lst
**   Save received trap data to file traps.lst.
**
** SnmpTool @testplan
**   Read command line arguments from the file "testplan".
**
** The command line arguments may be placed in a run command file and
** read by SnmpTool using the "@filename" command line syntax. The
** arguments in the file must be delimited using spaces or newline
** characters.  Lines beginning with a semicolon are ignored and all
** characters enclosed in double quotes are considered as one
** argument. Here are three examples of run command files:
** 
**   ; Sample SnmpTool command line input file
**   GET
**   111.11.11.1
**   public
**   1.2.0 1.3.6.1.4.1.77.1.1.15.0
** 
**   WALK 123.12.23.3 public
** 
**   SET 200.200.200.1 private
**   ; sysContact
**   1.4.0 2 "James D. Murray"
**   ; sysName
**   .1.3.6.1.2.1.1.5.0 2 PPRO200
**   ; sysLocation
**   1.5.0 2 "In the livingroom, just to the right of the piano"
**   ; oraMinAgentInteger
**   .1.3.6.1.4.1.2035.1.7.1.1.0 1 42
** 
**
** LICENSE: This program (source, object, and executable code) was
** created expressly for distribution in the book "Windows NT SNMP"
** by James D. Murray and published by O'Reilly and Associates
** (http://www.oreilly.com).  This program may be copied,
** distributed, and used by anyone and everyone in its original and
** unmodified form; the program source may also be used to derive
** other programs.  You MAY NOT, however, make any changes to this
** program and redistribute it. If you have bug reports and
** suggestions for enhancements, then please send them to the author.
**
** The updates of this program may be found at:
**
**      http://www.oreilly.com/catalog/winsnmp
**      ftp://ftp.oreilly.com/pub/examples/windows/winsnmp
**
** Revision History:
**     28 August 1997     1.00    Initial Release
*/
#include <conio.h>
#include <stdio.h>
#include <ctype.h>
#include <snmp.h>
#include <mgmtapi.h>
#include <iostream.h>
#include <winsock.h>

static const char * const gpCopyright = "Copyright 1997 by O'Reilly & Associates";

/* Local function prototypes */
BOOL ReceiveTrapsEventDriven( void );
BOOL ReceiveTrapsPollDriven( UINT, UINT );
BOOL ParseCommandFile( char *, int *, char ** );
BOOL CollectVarBinds( int *, char **, BYTE, RFC1157VarBindList * );
BOOL PollMib( char *, char *, INT, INT, UINT, UINT, RFC1157VarBindList * );
BOOL WalkMib( char *, char *, INT, INT, RFC1157VarBindList * );
SNMPAPI SendRequest( char *, char *, INT, INT, BYTE, RFC1157VarBindList *, AsnInteger *, AsnInteger *, BOOL );
void DisplayGetResponsePDU( UINT, UINT, RFC1157VarBindList *, BOOL );
void DisplayTrapResponsePDU( DWORD, AsnObjectIdentifier *, AsnNetworkAddress *, AsnInteger *, AsnInteger *, AsnTimeticks *, RFC1157VarBindList * );
void DisplayVarBinds(  RFC1157VarBindList  * );
char *AnyToStr( AsnObjectSyntax * );
void PrintOid( AsnObjectIdentifier * );
void PrintErrorMessage( char *, DWORD );
BOOL CtrlCHandler( DWORD );

#define OPER_GET     ASN_RFC1157_GETREQUEST
#define OPER_GETNEXT ASN_RFC1157_GETNEXTREQUEST
#define OPER_SET     ASN_RFC1157_SETREQUEST
#define OPER_WALK    1
#define OPER_POLL    2
#define OPER_TRAP    3

#define TIMEOUT           3000
#define RETRIES           3
#define POLLINGINTERVAL   3000U
#define NUMBEROFPOLLS     10U

/* Command line args and flags */
BYTE gbOperation       = 0;
INT  gnRequestTimeout  = TIMEOUT;
INT  gnRequestRetries  = RETRIES;
UINT guPollingInterval = POLLINGINTERVAL;
UINT guNumberOfPolls   = NUMBEROFPOLLS;
BOOL gfLongWalk        = FALSE;
BOOL gfPollForTraps    = FALSE;
BOOL gfGetNextPolling  = FALSE;

LPSNMP_MGR_SESSION ghMgmtSession = NULL;


static char *pTitle = "\nSnmpTool - Simple Network Management Protocol Tool for Win32\n";
static char *pAuthor = "Version 1.0 (28 Aug 1997); James D. Murray <jdm@oreilly.com>\n\n";
static char *pUsage = "\
Usage: SnmpTool [-rt] {get|getnext} agent community oid [oid] [...]\n\
       SnmpTool [-rt] set agent community oid type value [oid type value] [...]\n\
       SnmpTool [-inrtx] poll agent community oid [oid] [...]\n\
       SnmpTool [-lrt] walk agent community oid\n\
       SnmpTool [-inp] trap\n\
       SnmpTool @file\n\n\
       Options with arguments are:\n\
         -t <request time out in milliseconds>\n\
         -r <request retries attempts before timing out>\n\
         -i <polling interval in milliseconds>\n\
         -n <number of polling requests>\n\n\
       Valid \'type\' are:\n\
         1 = INTEGER            5 = Counter\n\
         2 = OCTET STRING       6 = Gauge\n\
         3 = OBJECT IDENTIFIER  7 = TimeTicks\n\
         4 = IpAddress          8 = Opaque\n";

/* Response ErrorStatus strings */
static char *szErrorStatus[] =
{ 
   "No Error",
   "Data is too big to return",
   "No such variable name",
   "Bad value sent",
   "Attempted writing to a read-only variable",
   "General error" 
};

/* Generic trap strings */
static char *szGenericTraps[] =
{ 
    "coldStart(0)",
    "warmStart(1)",
    "linkDown(2)",
    "linkUp(3)",
    "authenticationFailure(4)",
    "egpNeighborLoss(5)",
    "enterpriseSpecific(6)"
};



/*
** Main function.
**
** Returns: 0 on success, < 0 on error.
*/
int __cdecl main( 
  int    nArgc,     /* IN Command line argument count */
  char **pArgv )    /* IN Command line argument list */
{
    int                 argc;
    char                **argv;
    INT                 nFileArgc = 0;
    int                 nReturn = -1;
    char                szAgent[16];
    char                szCommunity[30];
    AsnInteger          uErrorStatus;
    AsnInteger          uErrorIndex;
    RFC1157VarBindList  VarBinds;

	WSAData		wsaData;

    /* Reroute Ctrl+C and Ctrl+Break */
    SetConsoleCtrlHandler( (PHANDLER_ROUTINE)CtrlCHandler, TRUE );

    argc = nArgc;
    argv = pArgv;

    /* Parse the command line arguments */
    if ( argc > 1 )
    {
        argv++; argc--;

        /*
        ** If the user has choosen to place the command line arguments in a file,
        ** we must parse the file and load the data into argc and argv variables
        ** as if the data were received from the command line.
        */
        if ( (*argv)[0] == '@' )
        {
            char szFile[81];
            strncpy( szFile, (*argv) + 1, sizeof( szFile ) - 1 );
            if ( ( nFileArgc = ParseCommandFile( szFile, &argc, argv ) ) == 0 )
            {
                printf( "SnmpTool: Failed to read command file %s\n", szFile );
                goto UsageError;
            }
        }

        /*
        ** Process option flags (if any)
        */
        while ( argv[0][0] == '-' || argv[0][0] == '/'  )
        {
            switch ( tolower( argv[0][1] ) )
            {
                case 'i':         /* Polling interval in milliseconds */
                    argv++; argc--;
                    if ( argc == 0 )
                        goto UsageError;
                    guPollingInterval = atoi( *argv );
                    break;
                case 'l':        /* Take a long walk through the MIB */
                    gfLongWalk = !gfLongWalk;
                    break;
                case 'n':         /* Number of polls */
                    argv++; argc--;
                    if ( argc == 0 )
                        goto UsageError;
                    guNumberOfPolls = atoi( *argv );
                    break;
                case 'p':        /* Use polled trap detection */
                    gfPollForTraps = !gfPollForTraps;
                    break;
                case 't':        /* Request time out in milliseconds */
                    argv++; argc--;
                    if ( argc == 0 )
                        goto UsageError;
                    gnRequestTimeout = atoi( *argv );
                    break;
                case 'r':        /* Number of request retries */
                    argv++; argc--;
                    if ( argc == 0 )
                        goto UsageError;
                    gnRequestRetries = atoi( *argv );
                    break;
                case 'x':        /* Use GETNEXT for polling */
                    gfGetNextPolling = !gfGetNextPolling;
                    break;
                default:
                    printf( "SnmpTool: Unknown flag: %c\n", argv[0][1] );
                    goto UsageError;
            }
            argv++; argc--;
            if ( argc == 0 )
                goto UsageError;
        }

        /*
        ** Get the operation to perform
        */
        if ( !stricmp( *argv, "get" ) )
        {
            gbOperation = OPER_GET;
            if ( argc < 4 )
                goto UsageError;
        }
        else
        if ( !stricmp( *argv, "getnext" ) )
        {
            gbOperation = OPER_GETNEXT;
            if ( argc < 4 )
                goto UsageError;
        }
        else
        if ( !stricmp( *argv, "set" ) )
        {
            gbOperation = OPER_SET;
            if ( argc < 6 || argc % 3 )
                goto UsageError;
        }
        else
        if ( !stricmp( *argv, "walk" ) )
        {
            gbOperation = OPER_WALK;
            if ( argc < 4 )
                goto UsageError;
            if ( argc > 5 )
            {
                puts( "SnmpTool: Specify only one starting OID for MIB walk." );
                goto UsageError;
            }
        }
        else
        if ( !stricmp( *argv, "poll" ) )
        {
            gbOperation = OPER_POLL;
            if ( argc < 4 )
                goto UsageError;
        }
        else
        if ( !stricmp( *argv, "trap" ) )
            gbOperation = OPER_TRAP;
        else
            goto UsageError;

        /*
        ** Perform the specified operation or listen for traps
        */
        fputs( pTitle, stderr );
        fputc( '\n', stderr );
        if ( gbOperation != OPER_TRAP )
        {
            argv++; argc--;
            strncpy( szAgent, *argv, sizeof( szAgent ) - 1 );
            argv++; argc--;
            strncpy( szCommunity, *argv, sizeof( szCommunity ) - 1 );

            argv++; argc--;
            if ( CollectVarBinds( &argc, argv, gbOperation, &VarBinds ) == TRUE )
            {
                nReturn = 0;
                if ( gbOperation == OPER_POLL )
                {
                    printf("\nPolling Interval: %u milliseconds, Number of Polls: %u\n\n", 
                      guPollingInterval, guNumberOfPolls );

                    if ( PollMib( szAgent, szCommunity, gnRequestTimeout, gnRequestRetries, 
                      guPollingInterval, guNumberOfPolls, &VarBinds ) == FALSE )
                        nReturn = -2;
                }
                else
                if ( gbOperation == OPER_WALK )
                {
                    printf("\nPerforming %s walk starting at OID ", 
                      gfLongWalk == TRUE ? "long" : "branch" );
                    PrintOid( &VarBinds.list->name );
                    puts("\n");

                     if ( WalkMib( szAgent, szCommunity, gnRequestTimeout, 
                      gnRequestRetries, &VarBinds ) == FALSE )
                        nReturn = -2;
                }
                else    /* Single Get, GetNext, or Set operation */
                {
                    if ( SendRequest( szAgent, szCommunity, gnRequestTimeout, gnRequestRetries, 
                      gbOperation, &VarBinds, &uErrorStatus, &uErrorIndex, FALSE ) == 0 )
                        DisplayGetResponsePDU( uErrorStatus, uErrorIndex, &VarBinds, TRUE );
                    else
                        nReturn = -2;
                }

                /* Free the allocated variable bindings */
                SnmpUtilVarBindListFree( &VarBinds );
            }
        }
        else
        {
            /* Check if we are running under Windows NT */
            OSVERSIONINFO vi;
            vi.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
            GetVersionEx( &vi );
            if ( vi.dwPlatformId == VER_PLATFORM_WIN32_NT )
            {
                nReturn = 0;
                if ( gfPollForTraps == FALSE )
                {
                    if ( ReceiveTrapsEventDriven() == FALSE )
                        nReturn = -2;
                }
                else
                {
                    if ( ReceiveTrapsPollDriven( guPollingInterval, guNumberOfPolls ) == FALSE )
                        nReturn = -2;
                }
            }
            else
            {
                puts( "\nSnmpTool: Trap detection is only supported under Windows NT" );
            }
        }
    }

UsageError:

    /* Close an open session */
    if ( ghMgmtSession )
        SnmpMgrClose( ghMgmtSession );

    /* Delete the pseudo command line arguments derived from a run command file */
    if ( nFileArgc )
    {
        for ( --nFileArgc; nFileArgc > 0; --nFileArgc )
            SnmpUtilMemFree( pArgv[nFileArgc] );
    }

    if ( nReturn == -1 )
    {
        fputs( pTitle, stderr );
        fputs( pAuthor, stderr );
        fputs( pUsage, stderr );
    }

    /* Enable Ctrl+C and Ctrl+Break */
    SetConsoleCtrlHandler( NULL, FALSE );

    return( nReturn );
}


/*
** Read the OIDs from the command line and assemble them into a
** variable binding list objects.
**
** Returns: 
*/
BOOL CollectVarBinds( 
  int                 *argc,          /* IN Command line argument count */
  char               **argv,          /* IN Command line argument list */
  BYTE                 bRequestType,  /* IN SNMP request type  */
  RFC1157VarBindList  *pVarBinds )    /* IN OUT Variable binding list */
{
    BOOL fResult = TRUE;
    UINT uVbIndex;
    AsnObjectIdentifier Oid;

    /*
    ** Insert formal parameter data validation code here
    */

    /* Create an empty varbinds list */
    pVarBinds->list = NULL;
    pVarBinds->len  = 0;

     while ( *argc )
    {
        pVarBinds->len++;

        if ( ( pVarBinds->list = 
          (RFC1157VarBind *)SnmpUtilMemReAlloc( pVarBinds->list, 
          sizeof( RFC1157VarBind ) * pVarBinds->len ) ) != NULL )
        {
            /* Initialize each varbind using the command line data */
            if ( SnmpMgrStrToOid( *argv, &Oid ) == FALSE )
            {
                  printf("SnmpTool: Error: Unknown or invalid OID specified (%s).\n", *argv );
                if ( GetLastError() != NO_ERROR )
                    PrintErrorMessage( "SnmpMgrStrToOid() Error", GetLastError() );
                fResult = FALSE;
                break;
            }                             
            uVbIndex = pVarBinds->len - 1;
            pVarBinds->list[uVbIndex].name = Oid;

            /* If we are performing a SET, then the next two args are the data type and value */
            if ( bRequestType == OPER_SET )
            {
                int type;
                (*argc)--;
                argv++;
                if ( argv == NULL )
                {
                    fResult = FALSE;
                    break;
                }
                type = atoi( *argv );
                (*argc)--;
                argv++;
                if ( argv == NULL )
                {
                    fResult = FALSE;
                    break;
                }

                switch( type )
                {
                    case 1: /* INTEGER */
                        pVarBinds->list[uVbIndex].value.asnType = ASN_INTEGER;
                        pVarBinds->list[uVbIndex].value.asnValue.number = atol(*argv);
                        break;
                    case 2: /* OCTET STRING */
                    {
                        char *p = *argv[0] == '\"' ? *argv + 1: *argv;

                        pVarBinds->list[uVbIndex].value.asnType = ASN_OCTETSTRING;
                        pVarBinds->list[uVbIndex].value.asnValue.string.length = lstrlen( p );

                        /* Strip the double quotes from the OCTET STRING value */
                        if ( p[pVarBinds->list[uVbIndex].value.asnValue.string.length-1] == '\"' )
                            pVarBinds->list[uVbIndex].value.asnValue.string.length--;

                        if ( pVarBinds->list[uVbIndex].value.asnValue.string.length > 0 )
                        {
                            pVarBinds->list[uVbIndex].value.asnValue.string.stream = (BYTE *)
                            SnmpUtilMemAlloc( pVarBinds->list[uVbIndex].value.asnValue.string.length*sizeof(BYTE) );
                            if ( pVarBinds->list[uVbIndex].value.asnValue.string.stream )
                            {
                                memcpy( (BYTE *)pVarBinds->list[uVbIndex].value.asnValue.string.stream,
                                  p,
                                  pVarBinds->list[uVbIndex].value.asnValue.string.length*sizeof(BYTE) );
                            }
                        }
                        else
                        {
                            fResult = FALSE;
                            break;
                        }
                        pVarBinds->list[uVbIndex].value.asnValue.string.dynamic = TRUE;
                        break;
                    }
                    case 3: /* OBJECT IDENTIFIER */
                    {
                        UINT i;
                        char *p = *argv[0] == '.' ? *argv + 1: *argv;

                        /* Count the number of sub-identifiers */
                        for ( i = 1; p = strchr( p, '.' ); i++, p++ )
                            ;

                        pVarBinds->list[uVbIndex].value.asnType = ASN_OBJECTIDENTIFIER;
                        pVarBinds->list[uVbIndex].value.asnValue.object.idLength = i;
                        if ( pVarBinds->list[uVbIndex].value.asnValue.object.idLength > 0 )
                        {
                            pVarBinds->list[uVbIndex].value.asnValue.object.ids = (UINT *)
                            SnmpUtilMemAlloc( pVarBinds->list[uVbIndex].value.asnValue.object.idLength*sizeof(UINT) );
                            if ( pVarBinds->list[uVbIndex].value.asnValue.object.ids )
                            {
                                i = 0;
                                p = *argv[0] == '.' ? *argv + 1: *argv;
                                for ( i = 0; i < pVarBinds->list[uVbIndex].value.asnValue.object.idLength; i++ )
                                {
                                    pVarBinds->list[uVbIndex].value.asnValue.object.ids[i] = atoi(p);
                                    if ( p = strchr(p, '.') )
                                        p++;
                                }
                            }
                        }
                        else
                        {
                            fResult = FALSE;
                            break;
                        }
                        break;
                    }
                    case 4: /* IpAddress */
                        pVarBinds->list[uVbIndex].value.asnType = ASN_RFC1155_IPADDRESS;
                        pVarBinds->list[uVbIndex].value.asnValue.address.length = 4;
                        pVarBinds->list[uVbIndex].value.asnValue.address.stream = (BYTE *)
                        SnmpUtilMemAlloc( pVarBinds->list[uVbIndex].value.asnValue.address.length*sizeof(BYTE) );

                        if ( pVarBinds->list[uVbIndex].value.asnValue.address.stream )
                        {
                            UINT i = 0;
                            char *p = *argv;
                            for ( i = 0; i < pVarBinds->list[uVbIndex].value.asnValue.address.length; i++ )
                            {
                                pVarBinds->list[uVbIndex].value.asnValue.address.stream[i] = atoi( p );
                                if ( p = strchr(p, '.') )
                                    p++;
                                else
                                    break;
                            }
                            if ( i != 3 )
                            {
                                fResult = FALSE;
                                break;
                            }
                        }
                        break;
                    case 5: /* Counter */
                        pVarBinds->list[uVbIndex].value.asnType = ASN_RFC1155_COUNTER;
                        pVarBinds->list[uVbIndex].value.asnValue.counter = atol(*argv);
                        break;
                    case 6: /* Gauge */
                        pVarBinds->list[uVbIndex].value.asnType = ASN_RFC1155_GAUGE;
                        pVarBinds->list[uVbIndex].value.asnValue.gauge = atol(*argv);
                        break;
                    case 7: /* TimeTicks */
                        pVarBinds->list[uVbIndex].value.asnType = ASN_RFC1155_TIMETICKS;
                        pVarBinds->list[uVbIndex].value.asnValue.ticks = atol(*argv);
                        break;
                    case 8: /* Opaque */ 
                        pVarBinds->list[uVbIndex].value.asnType = ASN_RFC1155_OPAQUE;
                        pVarBinds->list[uVbIndex].value.asnValue.arbitrary.length = lstrlen(*argv);
                        if ( pVarBinds->list[uVbIndex].value.asnValue.arbitrary.length > 0 )
                        {
                            pVarBinds->list[uVbIndex].value.asnValue.arbitrary.stream = (BYTE *)
                            SnmpUtilMemAlloc( pVarBinds->list[uVbIndex].value.asnValue.arbitrary.length*sizeof(BYTE) );
                            if ( pVarBinds->list[uVbIndex].value.asnValue.arbitrary.stream )
                            {
                                memcpy( (BYTE *)pVarBinds->list[uVbIndex].value.asnValue.arbitrary.stream,
                                  (char *)*argv,
                                  pVarBinds->list[uVbIndex].value.asnValue.arbitrary.length*sizeof(BYTE));
                            }
                        }
                        else
                        {
                            fResult = FALSE;
                            break;
                        }
                        pVarBinds->list[uVbIndex].value.asnValue.arbitrary.dynamic = TRUE;
                        break;
                    default:
                        fResult = FALSE;
                        break;
                }
                if ( fResult == FALSE )
                    break;
            }
            else
                pVarBinds->list[uVbIndex].value.asnType = ASN_NULL;
        }
        else
        {
            PrintErrorMessage( "SnmpUtilMemReAlloc() Error", GetLastError() );
            fResult = FALSE;
            break;
        }
        argv++; (*argc)--;
    }

    if ( fResult == FALSE )
        SnmpUtilVarBindListFree( pVarBinds );

    return( fResult );
}


/*
** Open a connection to an SNMP managed node, send a GetRequest, GetNextRequest,
** or SetRequest, wait for a GetResponse, and return the results.
**
** Returns: 0 on success; non-zero on failure.
*/
SNMPAPI SendRequest(
  char               *pszAgentAddress,   /* IN     IP address of agent */
  char               *pszCommunityName,  /* IN     Community name of agent */
  INT                 nTimeout,          /* IN     Length of time to wait for response from agent */
  INT                 nRetries,          /* IN     Number of retries before giving up if no response */
  BYTE                bRequestType,      /* IN     Type of SNMP operation to perform */
  RFC1157VarBindList *pVarBinds,         /* IN OUT Variable binding list */
  AsnInteger         *puErrorStatus,     /* OUT    Resulting ErrorStatus value */
  AsnInteger         *puErrorIndex,      /* OUT    Resulting ErrorIndex value */
  BOOL                fDontCloseHandle ) /* IN     Don't close the session handle before returning */
{
    SNMPAPI uResult = 0;

    /* Check the arguments */
    if ( pszAgentAddress == NULL ||
         pszCommunityName == NULL ||
         pVarBinds == NULL ||
         puErrorStatus == NULL ||
         puErrorIndex == NULL )
    {
        goto SendRequestFailed;
    }

    /* Check if the caller wants the previously opened handle closed */
    if ( fDontCloseHandle == FALSE )
    {
        if ( ghMgmtSession )
        {
            SnmpMgrClose( ghMgmtSession );
            ghMgmtSession = NULL;
        }
    }

    /* Open the session */
    if ( ghMgmtSession == NULL )
    {
        ghMgmtSession = SnmpMgrOpen( pszAgentAddress, pszCommunityName, nTimeout, nRetries );
        if ( ghMgmtSession == NULL )
        {
            PrintErrorMessage( "SnmpMgrOpen() Error", GetLastError() );
            uResult = -1;
            goto SendRequestFailed;
        }
    }

    /* Send the request to the agent and receive a response */
    if ( SnmpMgrRequest( ghMgmtSession, bRequestType, 
      pVarBinds, puErrorStatus, puErrorIndex ) == 0 )
    {
        uResult = GetLastError();
        fprintf( stderr, "\nSnmpTool: No response from agent (%s).\n",
           uResult == SNMP_MGMTAPI_SELECT_FDERRORS ? "Winsock socket descriptor error" :
          (uResult == SNMP_MGMTAPI_TIMEOUT ? "Request timed out" : "Unknown") );

        if ( uResult == SNMP_MGMTAPI_TIMEOUT )
            fputs( "\nVerify the IP address and community name are correct and try pinging the host.\n", stderr );
    }

    /* Close the session */
    if ( fDontCloseHandle == FALSE )
    {
        if ( SnmpMgrClose( ghMgmtSession ) == FALSE )
            PrintErrorMessage( "SnmpMgrClose() Error", GetLastError() );
        ghMgmtSession = NULL;
    }

SendRequestFailed:
    return( uResult );
}


/*
** Poll the MIB variables specified on the command line.
**
** Returns: TRUE if the polling concluded without error, otherwise FALSE.
*/
BOOL PollMib(
  char               *pszAgentAddress,    /* IN IP address of agent */
  char               *pszCommunityName,   /* IN Community name of agent */
  INT                 nTimeout,           /* IN Length of time to wait for response from agent */
  INT                 nRetries,           /* IN Number of retries before giving up if no response */
  UINT                uPollingInterval,   /* IN Milliseconds to sleep between polls */
  UINT                uNumberOfPolls,     /* IN Number of polls to perform */
  RFC1157VarBindList *pVarBinds )         /* IN OIDs to poll */
{
    BOOL fResult = TRUE;
    BYTE bOperation = OPER_GET;
    UINT uPollCount;
    UINT uErrorStatus;
    UINT uErrorIndex;
    RFC1157VarBindList *pVBTmp = NULL;

    /* Check the arguments */
    if ( pVarBinds == NULL || pszCommunityName == NULL || pszAgentAddress == NULL )
    {
        fResult = FALSE;
        goto PollMibFailed;
    }

    /* Check if user specified GetNext polling */
    if ( gfGetNextPolling == TRUE )
    {
        bOperation = OPER_GETNEXT;
        if ( pVBTmp = (RFC1157VarBindList *)SnmpUtilMemAlloc( sizeof(RFC1157VarBindList) ) )
            SnmpUtilVarBindListCpy( pVBTmp, pVarBinds );
        else
            goto PollMibFailed;
    }

    fprintf( stderr, "Polling %s using %s. Hit Escape to exit...\n\n", 
        pszAgentAddress, bOperation == OPER_GET ? "GET" : "GETNEXT" );

    /* Poll until finished */
    for ( uPollCount = 0; uPollCount < uNumberOfPolls; uPollCount++ )
    {
        printf( "Poll %u:\n", uPollCount+1 );
        if ( SendRequest( pszAgentAddress, pszCommunityName, nTimeout, nRetries, 
          bOperation, pVarBinds, &uErrorStatus, &uErrorIndex, TRUE ) == 0 )
        {
            DisplayGetResponsePDU( uErrorStatus, uErrorIndex, pVarBinds, TRUE );        

            /* Sleep until time for next poll */
            if ( uPollCount < uNumberOfPolls - 1 ) 
                   Sleep( uPollingInterval );

            /* Check if the user hit Escape */
            if ( _kbhit() )
                if ( _getch() == '\033' )
                    break;

            /* Copy the original varbinds over the varbinds 
               returned by the last GetNext operation */
            if ( pVBTmp )
                SnmpUtilVarBindListCpy( pVarBinds, pVBTmp );
        }
        else
        {
            fResult = FALSE;
            break;
        }
    }

    if ( pVBTmp )
        SnmpUtilVarBindListFree( pVBTmp );

PollMibFailed:
    return( fResult );
}


/*
** Perform a GetNext walk starting at the given OID.
**
** Returns: TRUE if the walk concluded without error, otherwise FALSE.
*/
BOOL WalkMib(
  char               *pszAgentAddress,  /* IN IP address of agent */
  char               *pszCommunityName, /* IN Community name of agent */
  INT                 nTimeout,         /* IN Length of time to wait for response from agent */
  INT                 nRetries,         /* IN Number of retries before giving up if no response */
  RFC1157VarBindList *pVarBinds )       /* IN OIDs to poll */
{
    BOOL fResult = TRUE;
    UINT uErrorStatus;
    UINT uErrorIndex;
    AsnObjectIdentifier rootOid;
    char *pString = NULL;

    /* Check the arguments */
    if ( pVarBinds == NULL || pszCommunityName == NULL || pszAgentAddress == NULL )
    {
        fResult = FALSE;
        goto WalkMibFailed;
    }

    /* Only a single start OID should be specified */
    if ( pVarBinds->len != 1 )
    {
        fResult = FALSE;
        goto WalkMibFailed;
    }

    /* Get the starting OID */
    SnmpUtilOidCpy( &rootOid, &pVarBinds->list->name );

    /* Loop until we can walk no more */
    while ( 1 )
    {
        if ( SendRequest( pszAgentAddress, pszCommunityName, nTimeout, nRetries, 
            OPER_GETNEXT, pVarBinds, &uErrorStatus, &uErrorIndex, TRUE ) == 0 ) 
        {
            /*
            ** Walk until we step into another branch or we bottom-out with a noSuchName error.
            ** If the "long walk" flag is set, then we only look for a noSuchName error.
            */
            if ( uErrorStatus == SNMP_ERRORSTATUS_NOERROR )
            {
                if ( gfLongWalk == FALSE )
                {
                    /* Check if we stepped into another branch */
                    if ( SnmpUtilOidNCmp( &pVarBinds->list[0].name, &rootOid, rootOid.idLength ) )
                    {
                        puts( "\nEnd of MIB branch" );
                        break;
                    }
                }
                DisplayGetResponsePDU( uErrorStatus, uErrorIndex, pVarBinds, FALSE );

                /* Check if the user hit Escape */
                if ( _kbhit() )
                    if ( _getch() == '\033' )
                        break;
            }
            else
                break;
        }
        else
            break;
    }

WalkMibFailed:
    return( fResult );
}


/*
** Receive traps using the event-driven mechanism of the SNMP Trap Service.
**
** Returns: TRUE if polling concluded without error, otherwise FALSE.
*/
BOOL ReceiveTrapsEventDriven( void )
{
    BOOL fResult = TRUE;
    HANDLE hTrapEvent;
    static char *pListening = "\nSnmpTool: Listening for traps. Hit Escape to exit...";

    /* Start listening for trap events */
    if ( SnmpMgrTrapListen( &hTrapEvent ) == FALSE )
    {
        DWORD dwError = GetLastError();
        if ( dwError != SNMP_MGMTAPI_TRAP_DUPINIT )
        {
            PrintErrorMessage("ERROR: SnmpMgrTrapListen: ", GetLastError() );
            goto ReceiveTrapsEventDrivenFailed;
            fResult = FALSE;
        }
    }
    else    /* Wait for a trap(s) */
    {
        DWORD dwTrapCount = 0;
        DWORD dwResult;

        fputs( pListening, stderr );

        while (1)
        {
            /* Receiving a trap will set an event signal */
            dwResult = WaitForSingleObject( hTrapEvent, 1000 );
            if ( dwResult != WAIT_FAILED )
            {
                if ( dwResult != WAIT_TIMEOUT )
                {    
                    AsnInteger          TrapGeneric;
                    AsnInteger          TrapSpecific;
                    AsnTimeticks        TrapTimeStamp;
                    AsnIPAddress        TrapIpAddress;
                    AsnObjectIdentifier TrapEnterprise;
                    RFC1157VarBindList  TrapVarBinds;

                    /* Initialize the variable bindings list */
                    TrapVarBinds.list = (RFC1157VarBind *) SnmpUtilMemAlloc( sizeof( RFC1157VarBind ) );
                    if ( TrapVarBinds.list == NULL )
                    {
                        PrintErrorMessage("ERROR: SnmpUtilMemAlloc: ", GetLastError() );
                        goto ReceiveTrapsEventDrivenFailed;
                        fResult = FALSE;
                    }
                        
                    /* Loop until all traps are read from the buffer */
                    while ( SnmpMgrGetTrap( &TrapEnterprise, &TrapIpAddress, 
                      &TrapGeneric, &TrapSpecific, &TrapTimeStamp, &TrapVarBinds ) == TRUE )
                    {
                        fputs( "\nReceived Trap!", stderr );
                        dwTrapCount++;
                        DisplayTrapResponsePDU( dwTrapCount, &TrapEnterprise, &TrapIpAddress, &TrapGeneric, 
                          &TrapSpecific, &TrapTimeStamp, &TrapVarBinds );

                        if ( TrapIpAddress.dynamic == TRUE )
                            SnmpUtilMemFree( TrapIpAddress.stream );
                        SnmpUtilOidFree( &TrapEnterprise );
                    }
                    SnmpUtilVarBindListFree( &TrapVarBinds );

                    /* Check if SnmpMgrGetTrap() returned an error */
                    if ( ( dwResult = GetLastError() ) != 0 )
                    {
                        if ( dwResult != SNMP_MGMTAPI_NOTRAPS )
                            PrintErrorMessage("ERROR: SnmpMgrGetTrap: ", dwResult );
                    }

                    if ( ResetEvent( hTrapEvent ) == 0 )
                        PrintErrorMessage("ERROR: ResetEvent: ", GetLastError() );

                    fputs( pListening, stderr );
                }

                /* Check if the user hit Escape */
                if ( _kbhit() )
                    if ( _getch() == '\033' )
                        break;
            }
            else
            {
                PrintErrorMessage( "ERROR: WaitForSingleObject: ", GetLastError() );
                goto ReceiveTrapsEventDrivenFailed;
                fResult = FALSE;
            }
        }
    }
ReceiveTrapsEventDrivenFailed:
    return( fResult );
}


/*
** Receive traps using a poll-driven mechanism.
**
** Returns: TRUE if polling concluded without error, otherwise FALSE.
*/
BOOL ReceiveTrapsPollDriven(
  UINT uPollingInterval,        /* IN Milliseconds to sleep between polls */
  UINT uNumberOfPolls )         /* IN Number of polls to perform */
{
    BOOL                fResult = TRUE;
    UINT                i;
    HANDLE              hTrapEvent;
    DWORD               dwTrapCount = 0;
    AsnObjectIdentifier TrapEnterprise;
    AsnNetworkAddress   TrapIpAddress;
    RFC1157VarBindList  TrapVarBinds;

    /* Start listening for trap events */
    if ( SnmpMgrTrapListen( &hTrapEvent ) == FALSE )
    {
        DWORD dwError = GetLastError();
        if ( dwError != SNMP_MGMTAPI_TRAP_DUPINIT )
        {
            PrintErrorMessage("ERROR: SnmpMgrTrapListen: ", GetLastError() );
            goto ReceiveTrapsPollDrivenFailed;
            fResult = FALSE;
        }
    }

    /* Initialize the variable bindings list */
    TrapVarBinds.list = (RFC1157VarBind *) SnmpUtilMemAlloc( 
      sizeof( RFC1157VarBind ) );

    /* Poll for traps */
    for ( i = 1; i <= uNumberOfPolls; i++ )
    {
        DWORD         dwError;
        BOOL         bReceivedTrap;
        AsnInteger   TrapGeneric;
        AsnInteger   TrapSpecific;
        AsnTimeticks TrapTimeStamp;

        fprintf( stderr, "SnmpTool: Polling for traps (poll %u). Hit Escape to exit...\n", i );
        do
        {
            /* Check if any traps were received */
            bReceivedTrap = SnmpMgrGetTrap( &TrapEnterprise, &TrapIpAddress, 
              &TrapGeneric, &TrapSpecific, &TrapTimeStamp, &TrapVarBinds );

            if ( bReceivedTrap == TRUE )
            {
                fputs( "Received Trap!\n", stderr );
                dwTrapCount++;
                DisplayTrapResponsePDU( dwTrapCount, &TrapEnterprise, &TrapIpAddress, 
                  &TrapGeneric, &TrapSpecific, &TrapTimeStamp, &TrapVarBinds );
            }

            /* Check if SnmpMgrGetTrap returned an error */
            dwError = GetLastError();
            if ( dwError != SNMP_MGMTAPI_NOTRAPS  && dwError != ERROR_SUCCESS )
            {
                /* A real error occurred */
                PrintErrorMessage( "SnmpTool: Trap polling error", dwError );
                fResult = FALSE;
                goto ReceiveTrapsPollDrivenFailed;
            }
        } while ( bReceivedTrap == TRUE );
  
        /* Sleep until time for next poll */
        if ( i < uNumberOfPolls ) 
            Sleep( uPollingInterval );

        /* Check if the user hit Escape */
        if ( _kbhit() )
            if ( _getch() == '\033' )
                break;
    }
    SnmpUtilVarBindFree( TrapVarBinds.list );

ReceiveTrapsPollDrivenFailed:
    return( fResult );
}


/*
** Display the information in a GetReponse PDU and its variable bindings.
*/
void DisplayGetResponsePDU(
  UINT                uErrorStatus,     /* IN Resulting ErrorStatus value */
  UINT                uErrorIndex,      /* IN Resulting ErrorIndex value */
  RFC1157VarBindList *pVarBinds,        /* IN Variable binding list */
  BOOL                fDisplayError )   /* IN Flag used to display SNMP error info */
{
    if ( fDisplayError == TRUE )
    {
        printf("  ErrorStatus: %u (%s)\n", 
          uErrorStatus,  
          uErrorStatus >= 0 && uErrorStatus <= 5 ? 
          szErrorStatus[uErrorStatus] : "Unknown");
        printf("   ErrorIndex: %u\n\n", uErrorIndex );
    }
    if ( pVarBinds )
        DisplayVarBinds( pVarBinds );
}


/*
** Display the information in a TrapReponse PDU and its variable bindings.
*/
void DisplayTrapResponsePDU(
  DWORD                dwTrapCount,     /* IN Count of traps received */
  AsnObjectIdentifier *pEnterprise,     /* IN TrapResponse PDU enterprises OID */
  AsnIPAddress        *pIPAddress,      /* IN TrapResponse PDU IpAddress value */
  AsnInteger          *plGenericTrap,   /* IN TrapResponse PDU generic trap value */
  AsnInteger          *plSpecificTrap,  /* IN TrapResponse PDU speciifc trap value */
  AsnTimeticks        *pdwTimeStamp,    /* IN TrapResponse PDU time stamp */
  RFC1157VarBindList  *pVarBinds )      /* IN Variable bindings list */
{
    printf( "\nTrap Received: %lu\n", dwTrapCount );
    printf( "   enterprise: " );
    PrintOid( pEnterprise );
    printf( "   agent-addr: %u.%u.%u.%u\n", pIPAddress->stream[0], 
      pIPAddress->stream[1], pIPAddress->stream[2], pIPAddress->stream[3] );
    printf( " generic-trap: %s\n", *plGenericTrap >= 0 && *plGenericTrap < 7 ? 
      szGenericTraps[*plGenericTrap] : "Unknown" );
    printf( "specific-trap: %lu\n", *plSpecificTrap );
    printf( "   time-stamp: %lu\n", *pdwTimeStamp );
    printf( "     varbinds: %u\n", pVarBinds->len );

    DisplayVarBinds( pVarBinds );
}


/*
** Display the contents of a RFC1157VarBindList object.
*/
void DisplayVarBinds(
  RFC1157VarBindList *pVarBinds )  /* IN Variable bindings list */
{
    UINT i, j;
    DWORD dwLength;
    char *pString = NULL;

    if ( pVarBinds == NULL )
        return;
	
	printf("\nI add something here ......\n");

    for ( i = 0; i < pVarBinds->len; i++ )
    {
        if ( gbOperation != OPER_WALK )
            printf( "    varbind %u: \n", i+1 );

        /* Name */ 
        SnmpMgrOidToStr( &pVarBinds->list[i].name, &pString );
        if ( pString )
        {
            printf( "         Name: %s \n", pString );
            SnmpUtilMemFree( pString );
            pString = NULL;
        }

        /* OID */
        printf( "          OID: " );
        for ( j = 0; j < pVarBinds->list[i].name.idLength; j++ )
        {
            printf( "%d%c", pVarBinds->list[i].name.ids[j],
                j < pVarBinds->list[i].name.idLength - 1 ? '.' : ' ' );
        }
        puts( "" );

        /* Type and Length */
        printf( "         Type: " );
        switch ( pVarBinds->list[i].value.asnType )
        {
            case ASN_INTEGER:
                puts( "INTEGER" );
                dwLength = ~0UL;
                break;
            case ASN_OCTETSTRING:
                puts( "OCTET STRING" );
                dwLength = pVarBinds->list[i].value.asnValue.string.length;              
                break;
            case ASN_NULL:
                puts( "NULL" );
                dwLength = ~0UL;
                break;
            case ASN_OBJECTIDENTIFIER:
                puts( "OBJECT IDENTIFIER" );
                dwLength = pVarBinds->list[i].value.asnValue.object.idLength;
                break;
            case ASN_SEQUENCE:
                puts( "SEQUENCE" );
                dwLength = pVarBinds->list[i].value.asnValue.sequence.length;
                break;
            case ASN_RFC1155_IPADDRESS:
                puts( "IpAddress" );
                dwLength = pVarBinds->list[i].value.asnValue.address.length;
                break;
            case ASN_RFC1155_COUNTER:
                puts( "Counter" );
                dwLength = ~0UL;
                break;
            case ASN_RFC1155_GAUGE:
                puts( "Gauge" );
                dwLength = ~0UL;
                break;
            case ASN_RFC1155_TIMETICKS:
                puts( "TimeTicks" );
                dwLength = ~0UL;
                break;
            case ASN_RFC1155_OPAQUE:
                puts( "Opaque" );
                dwLength = pVarBinds->list[i].value.asnValue.arbitrary.length;
                break;
            default:
                puts( "Unknown" );
                dwLength = 0;
                break;
        }

        /* Length */
        if ( dwLength != ~0 )
            printf( "       Length: %lu\n", dwLength );

        /* Value */
        printf( "        Value: " );
        if ( pString = AnyToStr( &pVarBinds->list[i].value ) )
        {
            printf( "%s \n", pString );
            SnmpUtilMemFree( pString );
            pString = NULL;
        }
        else
            puts( "<No Data>" );

        puts( "" );
    }
}


/*
** Convert an AsnAny value to a string.
**
** Returns: A string representation of the variable data.
*/
char * AnyToStr( 
  AsnObjectSyntax  *sAny )       /* IN  AsnObjectSyntax object containing value */
{
    DWORD dwValue = 0;
    UINT uLen = 0;
    BYTE *puData = 0;
    char *pString = NULL;

    switch ( sAny->asnType )
    {
        case ASN_INTEGER:    
            pString = (char *) SnmpUtilMemAlloc( 33 );
            if ( pString )
               ltoa( sAny->asnValue.number, pString, 10 );
            break;
        case ASN_RFC1155_COUNTER:
            dwValue = sAny->asnValue.counter;
            goto Uint;
        case ASN_RFC1155_GAUGE:
            dwValue = sAny->asnValue.gauge;
            goto Uint;
        case ASN_RFC1155_TIMETICKS:
            dwValue = sAny->asnValue.ticks;
Uint:
            pString = (char *) SnmpUtilMemAlloc( 33 );
            if ( pString )
               ultoa( dwValue, pString, 10 );
            break;
        case ASN_OCTETSTRING:   /* Same as ASN_RFC1213_DISPSTRING */
            uLen = sAny->asnValue.string.length;
            puData = sAny->asnValue.string.stream;
            goto String;
        case ASN_SEQUENCE:      /* Same as ASN_SEQUENCEOF */
            uLen = sAny->asnValue.sequence.length;
            puData = sAny->asnValue.sequence.stream;
            goto String;
        case ASN_RFC1155_IPADDRESS:
        {
            if ( sAny->asnValue.address.length )
            {
                UINT i;
                char szBuf[17];

                uLen = sAny->asnValue.address.length;
                puData = sAny->asnValue.address.stream;

                pString = (char *) SnmpUtilMemAlloc( uLen * 4 );
                if ( pString )
                {
                    pString[0] = '\0';
    
                    for ( i = 0; i < uLen; i++ )
                    {
                        lstrcat( pString, itoa( puData[i], szBuf, 10 ) );    
                        if ( i < uLen-1 )
                            lstrcat( pString, "." );
                    }
                }
            }
            else
                pString = NULL;
            break;
        }
        case ASN_RFC1155_OPAQUE:
            if ( sAny->asnValue.arbitrary.length )

            {
                uLen = sAny->asnValue.arbitrary.length;
                puData = sAny->asnValue.arbitrary.stream;
String:
                pString = (char *) SnmpUtilMemAlloc( uLen + 1 );
                if ( pString )
                {
                    if ( sAny->asnValue.arbitrary.length )
                        strncpy( pString, puData, uLen );
                }
            }
            else
                pString = NULL;
            break;
        case ASN_OBJECTIDENTIFIER:
        {
            if ( sAny->asnValue.object.idLength )
            {
                pString = (char *) SnmpUtilMemAlloc( sAny->asnValue.object.idLength * 5 );
                if ( pString )
                {
                    UINT i;
                    char szBuf[17];
                    for ( i = 0; i < sAny->asnValue.object.idLength; i++ )
                    {
                        lstrcat( pString, itoa( sAny->asnValue.object.ids[i], szBuf, 10 ) );    
                        if ( i < sAny->asnValue.object.idLength-1 )
                            lstrcat( pString, "." );
                    }
                }
            }
            else
                pString = NULL;
            break;
        }
        default:             /* Unrecognised data type */
            return( FALSE );
    }
    return( pString );
}


/*
** Display an AsnObjectIdentifier OID value as a string with a newline.
*/
void PrintOid( 
  AsnObjectIdentifier *pName )  /* IN  */
{
    if ( pName )
    {
        UINT i;
        char szBuf[8];
        char szOid[80];

        szOid[0] = '\0';

        for ( i = 0; i < pName->idLength; i++ )
        {
            wsprintf( szBuf, "%u.", pName->ids[i] );
            lstrcat( szOid, szBuf );
        }
        szOid[lstrlen(szOid)-1] = '\0';
    
        puts( szOid );
    }
}


/*
** Display an error string and the GetLastError message.
*/
void PrintErrorMessage( 
  char  *pString,            /* IN Error string to display */
  DWORD  dwError )            /* IN Value returned by GetLastError */
{
   char szMsg[128];

   if ( pString )
   {
      wsprintf( szMsg, "%s: %u", pString, dwError );
      puts( szMsg );
   }
   
   if ( FormatMessage( 
      FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_MAX_WIDTH_MASK,
      NULL, dwError,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
      (LPTSTR) &szMsg, sizeof( szMsg ) - 1, NULL ) != 0 )
   {
      puts( szMsg );
   }
}


/*
** Read in command line data from a file.
**
** This is a very simple aprser used to read the contents of the specified
** run command file and convert it to an equivalent argc/argv array.
** The arguments may be separated by whitespaces (including newlines) and
** lines beginning with a semicolon are considered comments. All characters
** appearing between double quotes are considered to be a single argument.
**
** Returns: The number of arguemnts processed if successful, otherwise 0.
*/
int ParseCommandFile( 
  char  *pFileName,     /* IN Run command file name */
  int   *pnArgc,        /* IN Command line argument count */
  char **pArgv )        /* IN Command line argument list */
{
    int  nReturn = 0;
    FILE *pFp;

    /* Check agruments */
    if ( pFileName == NULL || pnArgc == NULL || pArgv == NULL )
    {
        nReturn = 0;
        goto ParseCommandFileFailed;
    }

    *pnArgc = 0;

    /* Open the run command file */
    if ( ( pFp = (FILE *) fopen( pFileName, "rt" ) ) == (FILE *)NULL )
    {
        nReturn = 0;
        goto ParseCommandFileFailed;
    }

    /* Read in the run command file */
    while ( !feof( pFp ) )
    {
        int c, i;
        char szTmp[80];
        BOOL fDquoteIsOpen;

        fDquoteIsOpen = FALSE;
        
        /* Read in chars until we hit an EOLN or EOF */
        for ( i = 0; ( c = fgetc( pFp ) ) != EOF ; i++ )
        {
            /* At the end of an arg or the line */
            if ( ( c == ' ' && !fDquoteIsOpen ) || c == '\n' )
            {
                if ( i > 0 )                /* Not a blank line? */
                {
                    if ( szTmp[0] != ';' )    /* Line not a comment? */
                    {
                        szTmp[i] = '\0';
                        *pArgv = (char *)SnmpUtilMemAlloc( lstrlen( szTmp ) + 1 );
                        lstrcpy( *pArgv, szTmp );
                        (*pnArgc)++;
                        **pArgv++;
                    }
                    else      /* Eat and ignore a comment line */
                    {
                        while ( ( c = fgetc( pFp ) ) != EOF )
                            if ( c == '\n' )
                                break;
                    }
                }
                break;
            }
            else
            {
                /* Keep track of opened and closed double quotes */
                if ( c == '"' )
                    fDquoteIsOpen = !fDquoteIsOpen;

                szTmp[i] = (char)c;
            }
        }
    }
    fclose( pFp );

    nReturn = *pnArgc;

ParseCommandFileFailed:
    return( nReturn );
}


/*
** Handler function used to eat Cntl+C's.
*/
BOOL CtrlCHandler( DWORD dwEventType )
{
    /* Poke an ESCape into the console input to stop polling */
    _ungetch( '\033' );

    return( TRUE );
}



